import numpy as np
import math


def distance(p1, p2):
    Angs2Bohr = 1 / 5.2917726E-1
    p1 = [p * Angs2Bohr for p in p1]
    p2 = [p * Angs2Bohr for p in p2]
    return np.sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2 + (p1[2] - p2[2]) ** 2)


def number_of_sum(s, t):
    if (s+t)%2 == 0:
        return int(0.5*(s+t))
    else:
        return int(0.5*(s+t-1))


def combination(n, k):
    if k > n:
        return 0
    return math.factorial(n)/(math.factorial(k)*math.factorial(n - k))


def binomial_products(t, m, n, a, b):
    fn = 0
    for i in range(t+1):
        if m-i < 0 or n-t+i < 0:
            fn += 0
        else:
            fn += combination(m, i)*combination(n, t - i)*(a**(m-i)*b**(n-t+i))
    return fn


def double_factorial(n):
    if n == 0 or n == 1 or n == -1:
        return 1
    return n * double_factorial(n - 2)


def overlap_integral(orbital_type1, orbital_type2, coordinates_atom1, coordinates_atom2, a1, a2):

    orbital_type_to_vector = {'s': [0, 0, 0],
                              'px': [1, 0, 0],
                              'py': [0, 1, 0],
                              'pz': [0, 0, 1],
                              'dxx': [2, 0, 0],
                              'dxy': [1, 1, 0],
                              'dxz': [1, 0, 1],
                              'dyy': [0, 2, 0],
                              'dyz': [0, 1, 1],
                              'dzz': [0, 0, 2]}

    R12 = distance(coordinates_atom1, coordinates_atom2)
    orbital1 = orbital_type_to_vector[orbital_type1]
    orbital2 = orbital_type_to_vector[orbital_type2]
    exponential = ((np.pi/(a1 + a2))**(3/2)*np.exp(-(R12**2)*(a1*a2/(a1 + a2))))

    sum_of_binomial_products = 1
    Angs2Bohr = 1 / 5.2917726E-1
    for i in range(0,3):
        max_number_of_sum = number_of_sum(orbital1[i], orbital2[i])
        fxyz = 0
        for j in range(max_number_of_sum+1):
            a = Angs2Bohr*(coordinates_atom2[i] - coordinates_atom1[i])*a2/(a1 + a2)
            b = Angs2Bohr*(coordinates_atom1[i] - coordinates_atom2[i])*a1/(a1 + a2)
            fxyz += binomial_products(2 * j, orbital1[i], orbital2[i], a, b) * \
                    double_factorial(2 * j - 1) / ((2 ** j) * (a1 + a2) ** j)
        sum_of_binomial_products *= fxyz

    overlap = exponential*sum_of_binomial_products

    return overlap





import numpy as np
import copy
from huckelpy import tools


class AtomicOrbital():

    def __init__(self, atom, atomic_basis, position, orbital_type, eh_energy):

        self._atom = atom
        self._shell = atomic_basis
        self._position = position
        self._orbital_type = orbital_type
        self._energy = eh_energy
        self.linear_combination = None
        self.linear_combination_coefficients = None

    def get_linear_combination(self):
        if self.linear_combination is None:
            pure_basis = {'dz2': {'orbitals': ['dzz', 'dxx', 'dyy'], 'lcc': [1, -1/2, -1/2]},
                          'dx2-y2': {'orbitals': ['dxx', 'dyy'], 'lcc': [np.sqrt(3/4), -np.sqrt(3/4)]}}
            if self._orbital_type in pure_basis:
                self.linear_combination = pure_basis[self._orbital_type]['orbitals']
                self.linear_combination_coefficients = pure_basis[self._orbital_type]['lcc']
            else:
                self.linear_combination = [self._orbital_type]
                self.linear_combination_coefficients = [1]
        return self.linear_combination, self.linear_combination_coefficients

    def double_factorial(self, n):
        if n == 0 or n == 1 or n == -1:
            return 1
        return n * self.double_factorial(n - 2)

    def get_gaussian_norm(self):

        def double_factorial(n):
            if n == 0 or n == 1 or n == -1:
                return 1
            return n * double_factorial(n - 2)

        n1 = 0
        if self._orbital_type == 's':
            n1 = 1
        elif 'p' in self._orbital_type:
            n1 = 2
        elif 'd' in self._orbital_type:
            n1 = 3
        elif 'f' in self._orbital_type:
            n1 = 4

        norm_vector = []
        for alpha in self.ao_orbital_shell['p_exponents']:
            norm_vector.append((alpha ** ((2 * n1 + 1) / 4)) * (2 ** (n1 + 1)) / (
                    (double_factorial(2 * n1 - 1)) * (2 * np.pi) ** (-1 / 4)))
        return norm_vector

    def get_real_spherical_harmonic(self):
        if 's' in self._orbital_type:
            return np.sqrt(1 / (4 * np.pi))
        elif 'p' in self._orbital_type:
            return np.sqrt(3 / (4 * np.pi))
        elif 'd' in self._orbital_type:
            return np.sqrt(15 / (4 * np.pi))
        elif 'f' in self._orbital_type:
            return np.sqrt(105 / (4 * np.pi))

    def get_atomic_orbital_coeff(self):
        if 'p' in self._orbital_type:
            coefficients = copy.deepcopy(self.ao_orbital_shell['p_con_coefficients'])
        else:
            coefficients = copy.deepcopy(self.ao_orbital_shell['con_coefficients'])
        return coefficients

    def get_atomic_orbital_exponents(self):
        return copy.deepcopy(self.ao_orbital_shell['p_exponents'])

    def get_ao_row(self):

        if 'd' in self._orbital_type:
            row = str(tools.get_element_row(self.atom) - 1)
        elif 'f' in self._orbital_type:
            row = str(tools.get_element_row(self.atom) - 2)
        else:
            row = str(tools.get_element_row(self.atom))
        return row

    @property
    def ao_coefficients(self):
        return self.get_atomic_orbital_coeff()

    @property
    def ao_exponents(self):
        return self.get_atomic_orbital_exponents()

    @property
    def gaussian_norm(self):
        return self.get_gaussian_norm()

    def real_spherical_harmonic(self):
        return self.get_real_spherical_harmonic()

    @property
    def ao_orbital_shell(self):
        return self._shell

    @property
    def ao_linear_combination(self):
        return self.get_linear_combination()

    @property
    def position(self):
        return self._position

    @property
    def atom(self):
        return self._atom

    @property
    def type(self):
        return self._orbital_type

    @property
    def energy(self):
        return self._energy

    @property
    def row(self):
        return self.get_ao_row()
